@@ -54,7 +54,7 @@
 #include "strbuf.h"
 #include "util.h"
 #include "variable.h"
-#include "wfnmatch.h"
+#include "xfnmatch.h"
 #include "yash.h"
 
 
@@ -236,7 +236,7 @@ inline bool return_pending(void)
 /* Returns true iff we're breaking/continuing/returning now. */
 bool need_break(void)
 {
-    return execinfo.breakcount > 0 || return_pending();
+    return execinfo.breakcount > 0 || return_pending() || is_interrupted();
 }
 
 
@@ -266,7 +266,8 @@ void exec_pipelines(const pipeline_T *p, bool finally_exit)
 	bool savesee = supresserrexit;
 	supresserrexit |= p->pl_neg || p->next;
 
-	bool self = finally_exit && !p->next && !p->pl_neg && !any_trap_set;
+	bool self = finally_exit && !doing_job_control_now
+	    && !p->next && !p->pl_neg && !any_trap_set;
 	exec_commands(p->pl_commands, self ? execself : execnormal);
 	if (p->pl_neg)
 	    laststatus = !laststatus;
@@ -280,8 +281,6 @@ void exec_pipelines(const pipeline_T *p, bool finally_exit)
 /* Executes pipelines asynchronously. */
 void exec_pipelines_async(const pipeline_T *p)
 {
-    assert(!need_break());
-
     if (!p->next && !p->pl_neg) {
 	exec_commands(p->pl_commands, execasync);
     } else {
@@ -357,7 +356,6 @@ bool exec_condition(const and_or_T *c)
 void exec_for(const command_T *c, bool finally_exit)
 {
     assert(c->c_type == CT_FOR);
-    assert(!need_break());
     execinfo.loopnest++;
 
     int count;
@@ -372,7 +370,7 @@ void exec_for(const command_T *c, bool finally_exit)
 	struct get_variable v = get_variable(L"@");
 	assert(v.type == GV_ARRAY && v.values != NULL);
 	words = duparray(v.values, copyaswcs);
-	count = plcount(words);
+	count = (int) v.count;
     }
 
 #define CHECK_LOOP                                   \
@@ -384,6 +382,8 @@ void exec_for(const command_T *c, bool finally_exit)
 	continue;                                    \
     } else if (execinfo.exception != ee_none) {      \
 	goto done;                                   \
+    } else if (is_interrupted()) {                   \
+	goto done;                                   \
     } else (void) 0
 
     int i;
@@ -392,6 +392,9 @@ void exec_for(const command_T *c, bool finally_exit)
 		    posixly_correct ? SCOPE_GLOBAL : SCOPE_LOCAL, false))
 	    goto done;
 	exec_and_or_lists(c->c_forcmds, finally_exit && i + 1 == count);
+
+	if (!c->c_forcmds)
+	    handle_signals();
 	CHECK_LOOP;
     }
 
@@ -414,14 +417,23 @@ finish:
 void exec_while(const command_T *c, bool finally_exit)
 {
     assert(c->c_type == CT_WHILE);
-    assert(!need_break());
     execinfo.loopnest++;
 
     int status = Exit_SUCCESS;
-    while (exec_condition(c->c_whlcond) == c->c_whltype) {
+    for (;;) {
+	bool cond = exec_condition(c->c_whlcond);
+
+	if (!c->c_whlcond)
+	    handle_signals();
 	CHECK_LOOP;
+
+	if (cond != c->c_whltype)
+	    break;
 	exec_and_or_lists(c->c_whlcmds, false);
 	status = laststatus;
+
+	if (!c->c_whlcmds)
+	    handle_signals();
 	CHECK_LOOP;
     }
 
@@ -437,7 +449,6 @@ done:
 void exec_case(const command_T *c, bool finally_exit)
 {
     assert(c->c_type == CT_CASE);
-    assert(!need_break());
 
     wchar_t *word = expand_single(c->c_casword, tt_single);
     if (!word)
@@ -449,9 +460,16 @@ void exec_case(const command_T *c, bool finally_exit)
 	    wchar_t *pattern = expand_single(*pats, tt_single);
 	    if (!pattern)
 		goto fail;
-	    size_t match = wfnmatch(pattern, word, 0, WFNM_WHOLE);
+
+	    xfnmatch_T *xfnm = xfnm_compile(
+		    pattern, XFNM_HEADONLY | XFNM_TAILONLY);
 	    free(pattern);
-	    if (match != WFNM_NOMATCH && match != WFNM_ERROR) {
+	    if (!xfnm)
+		continue;
+
+	    bool match = (xfnm_wmatch(xfnm, word).start != (size_t) -1);
+	    xfnm_free(xfnm);
+	    if (match) {
 		if (ci->ci_commands) {
 		    exec_and_or_lists(ci->ci_commands, finally_exit);
 		    goto done;
@@ -536,11 +554,9 @@ void exec_commands(command_T *c, exec_T type)
     pid_t pgid;
     command_T *cc;
     job_T *job;
-    process_T *ps;
+    process_T *ps, *pp;
     pipeinfo_T pinfo = PIPEINFO_INIT;
-    commandtype_T lasttype = lasttype;
-
-    assert(!need_break());
+    commandtype_T lasttype;
 
     /* count the number of commands */
     count = 0;
@@ -552,31 +568,29 @@ void exec_commands(command_T *c, exec_T type)
     ps = job->j_procs;
 
     /* execute commands */
-    pgid = 0;
-    cc = c;
-    for (size_t i = 0; i < count; i++) {
+    pgid = 0, cc = c, pp = ps;
+    do {
 	pid_t pid;
 
 	lasttype = cc->c_type;
-	next_pipe(&pinfo, i < count - 1);
+	next_pipe(&pinfo, cc->next != NULL);
 	pid = exec_process(cc,
-		(type == execself && i < count - 1) ? execnormal : type,
+		(type == execself && cc->next != NULL) ? execnormal : type,
 		&pinfo,
 		pgid);
-	ps[i].pr_pid = pid;
+	pp->pr_pid = pid;
 	if (pid) {
-	    ps[i].pr_status = JS_RUNNING;
-	    ps[i].pr_statuscode = 0;
+	    pp->pr_status = JS_RUNNING;
+	    pp->pr_statuscode = 0;
 	} else {
-	    ps[i].pr_status = JS_DONE;
-	    ps[i].pr_statuscode = laststatus;
+	    pp->pr_status = JS_DONE;
+	    pp->pr_statuscode = laststatus;
 	}
-	ps[i].pr_name = NULL;   /* name is given later */
+	pp->pr_name = NULL;   /* name is given later */
 	if (pgid == 0)
 	    pgid = pid;
-	cc = cc->next;
-    }
-    assert(cc == NULL);
+	cc = cc->next, pp++;
+    } while (cc != NULL);
     assert(type != execself); /* `exec_process' doesn't return for `execself' */
     assert(pinfo.pi_tonextfds[PIDX_IN] < 0);
     if (pinfo.pi_fromprevfd >= 0)
@@ -584,42 +598,37 @@ void exec_commands(command_T *c, exec_T type)
     if (pinfo.pi_tonextfds[PIDX_OUT] >= 0)
 	xclose(pinfo.pi_tonextfds[PIDX_OUT]);  /* close leftover pipes */
 
-    if (pgid == 0) {  /* no more things to do if didn't fork */
+    if (pgid == 0) {
+	/* no more things to do if didn't fork */
 	free(job);
-	goto done;
-    }
-
-    job->j_pgid = doing_job_control_now ? pgid : 0;
-    job->j_status = JS_RUNNING;
-    job->j_statuschanged = true;
-    job->j_nonotify = false;
-    job->j_pcount = count;
-    set_active_job(job);
-    if (type == execnormal) {   /* wait for job to finish */
-	wait_for_job(ACTIVE_JOBNO, doing_job_control_now, false, false);
-	laststatus = calc_status_of_job(job);
-	if (job->j_status == JS_DONE) {
+    } else {
+	job->j_pgid = doing_job_control_now ? pgid : 0;
+	job->j_status = JS_RUNNING;
+	job->j_statuschanged = true;
+	job->j_nonotify = false;
+	job->j_pcount = count;
+	set_active_job(job);
+	if (type == execnormal) {
+	    wait_for_job(ACTIVE_JOBNO, doing_job_control_now, false, false);
+	    if (doing_job_control_now)
+		put_foreground(shell_pgid);
+	    laststatus = calc_status_of_job(job);
+	} else {
+	    assert(type == execasync);
+	    laststatus = Exit_SUCCESS;
+	    lastasyncpid = ps[count - 1].pr_pid;
+	}
+	if (job->j_status != JS_DONE) {
+	    for (cc = c, pp = ps; cc != NULL; cc = cc->next, pp++)
+		pp->pr_name = command_to_wcs(cc);
+	    add_job(type == execnormal || shopt_curasync);
+	} else {
 	    notify_signaled_job(ACTIVE_JOBNO);
 	    remove_job(ACTIVE_JOBNO);
-	    goto done;
 	}
-    } else {
-	laststatus = Exit_SUCCESS;
-	lastasyncpid = ps[count - 1].pr_pid;
     }
 
-    /* add the new job to the job list */
-    cc = c;
-    for (size_t i = 0; i < count; i++) {  /* name processes */
-	ps[i].pr_name = command_to_wcs(cc);
-	cc = cc->next;
-    }
-    add_job(type == execnormal || shopt_curasync);
-
-done:
-    if (doing_job_control_now)
-	put_foreground(shell_pgid); /* put the shell in the foreground */
-    handle_traps();
+    handle_signals();
     if (shopt_errexit && !supresserrexit && laststatus != Exit_SUCCESS
 	    && lasttype == CT_SIMPLE)
 	exit_shell();
@@ -801,20 +810,18 @@ done:
  * `sigtype' specifies settings of signals in the child.
  * `sigtype' is a bitwise OR of the followings:
  *   t_quitint: SIGQUIT & SIGINT are ignored if the parent's job control is off
- *   t_tstp: SIGTSTP is ignored if the parent is interactive
+ *   t_tstp: SIGTSTP is ignored if the parent is job-controlling
  *   t_leave: don't clear traps and shellfds. This option should be used only if
  *          the shell is going to `exec' to an extenal program.
  * Returns the return value of `fork'. */
 pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 {
-    sigset_t all, savemask;
-    bool sigblock = is_interactive_now || doing_job_control_now
-	|| (sigtype & (t_quitint | t_tstp));
-    if (sigblock) {
+    sigset_t savemask;
+    if (sigtype & (t_quitint | t_tstp)) {
+	sigset_t all;
 	sigfillset(&all);
 	sigemptyset(&savemask);
-	if (sigprocmask(SIG_BLOCK, &all, &savemask) < 0)
-	    xerror(errno, "sigprocmask"), sigblock = false;
+	sigprocmask(SIG_BLOCK, &all, &savemask);
     }
 
     pid_t cpid = fork();
@@ -828,39 +835,33 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 	    if (doing_job_control_now && pgid >= 0)
 		setpgid(cpid, pgid);
 	}
+	if (sigtype & (t_quitint | t_tstp))
+	    sigprocmask(SIG_SETMASK, &savemask, NULL);
     } else {
 	/* child process */
-	bool save_is_interactive_now = is_interactive_now;
 	bool save_doing_job_control_now = doing_job_control_now;
 	if (save_doing_job_control_now && pgid >= 0) {
 	    setpgid(0, pgid);
-	    if (fg) {
-		if (pgid == 0)
-		    pgid = getpgrp();
+	    if (pgid == 0)
+		pgid = getpgrp();
+	    shell_pgid = pgid;
+	    if (fg)
 		put_foreground(pgid);
-	    }
 	}
-	forget_initial_pgid();
-	neglect_all_jobs();
-	restore_job_signals();
-	restore_interactive_signals();
-	if (sigtype & t_leave) {
-	    clear_shellfds(true);
-	} else {
-	    clear_shellfds(false);
-	    clear_traps();
-	}
-	is_interactive_now = false;
 	if (sigtype & t_quitint)
 	    if (posixly_correct || !save_doing_job_control_now)
 		ignore_sigquit_and_sigint();
 	if (sigtype & t_tstp)
-	    if (save_is_interactive_now)
+	    if (save_doing_job_control_now)
 		ignore_sigtstp();
+	restore_signals(sigtype & t_leave);  /* signal mask is restored here */
+	clear_shellfds(sigtype & t_leave);
+	if (!(sigtype & t_leave)) {
+	    clear_traps();
+	    neglect_all_jobs();
+	}
+	is_interactive_now = false;
     }
-    if (sigblock)
-	if (sigprocmask(SIG_SETMASK, &savemask, NULL) < 0)
-	    xerror(errno, "sigprocmask");
     return cpid;
 }
 
@@ -1059,7 +1060,7 @@ void exec_simple_command(
 	    }
 	    mbsargv[argc] = NULL;
 
-	    restore_all_signals();
+	    restore_signals(true);
 	    xexecv(ci->ci_path, mbsargv);
 	    int errno_ = errno;
 	    if (errno_ != ENOEXEC) {
@@ -1076,7 +1077,6 @@ void exec_simple_command(
 	    for (int i = 1; i < argc; i++)
 		free(mbsargv[i]);
 	    laststatus = (errno_ == ENOENT) ? Exit_NOTFOUND : Exit_NOEXEC;
-	    init_signal();
 	    set_signals();
 	}
 	break;
@@ -1242,7 +1242,6 @@ wchar_t *exec_command_substitution(const wchar_t *code)
 	FILE *f;
 
 	xclose(pipefd[PIDX_OUT]);
-	set_nonblocking(pipefd[PIDX_IN]);
 	f = fdopen(pipefd[PIDX_IN], "r");
 	if (!f) {
 	    xerror(errno, Ngt("cannot open pipe for command substitution"));
@@ -1251,38 +1250,14 @@ wchar_t *exec_command_substitution(const wchar_t *code)
 	    return NULL;
 	}
 
-	/* make the child process the active job */
-	job_T *job = xmalloc(sizeof *job + sizeof *job->j_procs);
-	job->j_pgid = 0;
-	job->j_status = JS_RUNNING;
-	job->j_statuschanged = false;
-	job->j_nonotify = false;
-	job->j_pcount = 1;
-	job->j_procs[0].pr_pid = cpid;
-	job->j_procs[0].pr_status = JS_RUNNING;
-	job->j_procs[0].pr_statuscode = 0;
-	job->j_procs[0].pr_name = NULL;
-	set_active_job(job);
-
 	/* read output from the command */
 	xwcsbuf_T buf;
 	wb_init(&buf);
 	for (;;) {
 	    wint_t c = fgetwc(f);
 	    if (c == WEOF) {
-		if (feof(f))
+		if (feof(f) || errno != EINTR)
 		    break;
-		assert(ferror(f));
-		switch (errno) {
-		    case EINTR:
-		    case EAGAIN:
-#if EAGAIN != EWOULDBLOCK
-		    case EWOULDBLOCK:
-#endif
-			wait_for_input(pipefd[PIDX_IN], false, -1);
-			continue;
-		}
-		break;
 	    } else {
 		wb_wccat(&buf, c);
 	    }
@@ -1290,9 +1265,10 @@ wchar_t *exec_command_substitution(const wchar_t *code)
 	fclose(f);
 
 	/* wait for the child to finish */
-	wait_for_job(ACTIVE_JOBNO, false, false, false);
-	lastcmdsubstatus = calc_status_of_job(job);
-	remove_job(ACTIVE_JOBNO);
+	int savelaststatus = laststatus;
+	wait_for_child(cpid, 0, false);
+	lastcmdsubstatus = laststatus;
+	laststatus = savelaststatus;
 
 	/* trim trailing newlines and return */
 	while (buf.length > 0 && buf.contents[buf.length - 1] == L'\n')
@@ -1633,7 +1609,7 @@ int dot_builtin(int argc, void **argv)
     FILE *f = fopen(path, "r");
     if (path != mbsfilename)
 	free(path);
-    f = reopen_with_shellfd(f, "r");
+    f = reopen_with_shellfd(f, "r", true);
     if (!f) {
 	xerror(errno, Ngt("cannot open `%s'"), mbsfilename);
 	free(mbsfilename);
@@ -1777,8 +1753,7 @@ int exec_builtin_2(int argc, void **argv, const wchar_t *as, bool clear)
 	}
     }
 
-    finalize_shell();
-    restore_all_signals();
+    restore_signals(true);
     xexecve(commandpath, args, envs ? envs : environ);
     if (errno != ENOEXEC) {
 	if (errno == EACCES && is_directory(commandpath))
@@ -1793,9 +1768,7 @@ int exec_builtin_2(int argc, void **argv, const wchar_t *as, bool clear)
     }
     if (posixly_correct || !is_interactive_now)
 	exit(Exit_NOEXEC);
-    init_signal();
     set_signals();
-    reinitialize_shell();
     err = Exit_NOEXEC;
 
     recfree((void **) envs, free);
