@@ -201,6 +201,8 @@ void update_time(void)
 void set_histsize(unsigned newsize)
 {
     assert(newsize <= MAX_HISTSIZE);
+    if (newsize == 0)
+	newsize = 1;
     histsize = newsize;
 
     max_number = MIN_MAX_NUMBER;
@@ -682,6 +684,7 @@ void update_history(bool refresh)
 	/* The revision has been changed. Re-read everything. */
 	clear_all_entries();
 	clear_histfile_pids();
+	add_histfile_pid(shell_pid);
 	histfilerev = rev;
     }
     read_history(histfile);
@@ -1439,7 +1442,7 @@ int fc_edit_and_exec_entries(
 	    xerror(0, Ngt("editor returned non-zero status"));
 	    f = NULL;
 	} else {
-	    f = reopen_with_shellfd(fopen(temp, "r"), "r");
+	    f = reopen_with_shellfd(fopen(temp, "r"), "r", false);
 	    if (f == NULL)
 		xerror(errno, Ngt("cannot read command from `%s'"), temp);
 	}
