@@ -42,6 +42,9 @@
 #include "strbuf.h"
 #include "util.h"
 #include "yash.h"
+#if YASH_ENABLE_LINEEDIT && !defined(FG_DONT_SAVE_TERMINAL)
+# include "lineedit/terminfo.h"
+#endif
 
 
 static inline job_T *get_job(size_t jobnumber)
@@ -358,19 +361,29 @@ out_of_loop:
  * This function returns immediately if the job is already finished/stopped or
  * is not a child of this shell process.
  * Returns the signal number if interrupted, or zero if successful. */
+/* In most cases, you should call `put_foreground' to bring the shell back to
+ * foreground after calling `wait_for_job' if `doing_job_control_now' is true.
+ */
 int wait_for_job(size_t jobnumber, bool return_on_stop,
 	bool interruptible, bool return_on_trap)
 {
     int signum = 0;
     job_T *job = joblist.contents[jobnumber];
-    bool savenonotify = job->j_nonotify;
 
-    job->j_nonotify = true;
-    while (job->j_pgid >= 0 && job->j_status != JS_DONE
-	    && (!return_on_stop || job->j_status != JS_STOPPED)) {
-	signum = wait_for_sigchld(interruptible, return_on_trap);
+    if (job->j_pgid >= 0) {
+	bool savenonotify = job->j_nonotify;
+	job->j_nonotify = true;
+	for (;;) {
+	    if (job->j_status == JS_DONE)
+		break;
+	    if (return_on_stop && job->j_status == JS_STOPPED)
+		break;
+	    signum = wait_for_sigchld(interruptible, return_on_trap);
+	    if (signum != 0)
+		break;
+	}
+	job->j_nonotify = savenonotify;
     }
-    job->j_nonotify = savenonotify;
     return signum;
 }
 
@@ -416,30 +429,29 @@ wchar_t **wait_for_child(pid_t cpid, pid_t cpgid, bool return_on_stop)
     }
 }
 
-/* Sends the specified signal to the specified job.
- * See `get_jobnumber_from_name' for the format of `jobname'.
- * The return value is:
- *  - 0 if the signal is successfully sent,
- *  - 1 if there is no job for `jobname',
- *  - 2 if more than one jobs are found for `jobname',
- *  - 3 if the specified job is not job-controlled, or
- *  - 4 if `kill' returned an error (errno is set). */
-int send_signal_to_job(int signum, const wchar_t *jobname)
+/* Returns the process group ID of the specified job.
+ * If no valid job is found, an error message is printed and -1 is returned.
+ * `jobname' may have a preceding '%' sign. */
+pid_t get_job_pgid(const wchar_t *jobname)
 {
-    size_t jobnumber = get_jobnumber_from_name(jobname);
-    if (jobnumber == 0)
-	return 1;
-    if (jobnumber >= joblist.length)
-	return 2;
+    size_t jobnumber = get_jobnumber_from_name(
+	    jobname[0] == L'%' ? jobname + 1 : jobname);
+    const job_T *job;
 
-    job_T *job = get_job(jobnumber);
-    if (!job)
-	return 1;
-    if (job->j_pgid == 0)
-	return 3;
-    if (kill(-job->j_pgid, signum) < 0)
-	return 4;
-    return 0;
+    if (jobnumber >= joblist.length) {
+	xerror(0, Ngt("%ls: ambiguous job specification"), jobname);
+	return -1;
+    } else if (jobnumber == 0
+	    || (job = joblist.contents[jobnumber]) == NULL
+	    || job->j_pgid < 0) {
+	xerror(0, Ngt("%ls: no such job"), jobname);
+	return -1;
+    } else if (job->j_pgid == 0) {
+	xerror(0, Ngt("%ls: not job-controlled job"), jobname);
+	return -1;
+    } else {
+	return job->j_pgid;
+    }
 }
 
 /* Puts the specified process group in the foreground.
@@ -447,7 +459,6 @@ int send_signal_to_job(int signum, const wchar_t *jobname)
  * true. */
 void put_foreground(pid_t pgrp)
 {
-    bool saveok;
     sigset_t blockss, savess;
 
     assert(doing_job_control_now);
@@ -456,12 +467,45 @@ void put_foreground(pid_t pgrp)
     sigemptyset(&blockss);
     sigaddset(&blockss, SIGTTOU);
     sigemptyset(&savess);
-    saveok = sigprocmask(SIG_BLOCK, &blockss, &savess) >= 0;
+    sigprocmask(SIG_BLOCK, &blockss, &savess);
 
     tcsetpgrp(ttyfd, pgrp);
 
-    if (saveok)
-	sigprocmask(SIG_SETMASK, &savess, NULL);
+    sigprocmask(SIG_SETMASK, &savess, NULL);
+}
+
+/* Ensures the current shell process is in the foreground.
+ * The shell process is stopped by SIGTTOU until it is put in the foreground.
+ * This function requires `doing_job_control_now' to be true. */
+/* This function prevents the job-control shell from mangling the terminal while
+ * another shell is using it. */
+void ensure_foreground(void)
+{
+    /* This function calls `tcsetpgrp' with the default SIGTTOU handler. If the
+     * shell is in the background, it will receive SIGTTOU and is stopped until
+     * continued in the foreground. */
+
+    struct sigaction dflsa, savesa;
+    sigset_t blockss, savess;
+
+    assert(doing_job_control_now);
+    assert(shell_pgid > 0);
+
+    dflsa.sa_handler = SIG_DFL;
+    dflsa.sa_flags = 0;
+    sigemptyset(&dflsa.sa_mask);
+    sigemptyset(&savesa.sa_mask);
+    sigaction(SIGTTOU, &dflsa, &savesa);
+
+    sigemptyset(&blockss);
+    sigaddset(&blockss, SIGTTOU);
+    sigemptyset(&savess);
+    sigprocmask(SIG_UNBLOCK, &blockss, &savess);
+
+    tcsetpgrp(ttyfd, shell_pgid);
+
+    sigprocmask(SIG_SETMASK, &savess, NULL);
+    sigaction(SIGTTOU, &savesa, NULL);
 }
 
 /* Computes the exit status from the status code returned by `waitpid'. */
@@ -664,7 +708,9 @@ void print_job_status_all(bool changedonly, bool verbose, FILE *f)
 }
 
 /* If the shell is interactive and the specified job has been killed by a
- * signal other than SIGINT or SIGPIPE, prints a notification to stderr. */
+ * signal other than SIGPIPE, prints a notification to stderr.
+ * If the signal is SIGINT, only a single newline is printed to stderr and the
+ * shell is flagged as interrupted. */
 void notify_signaled_job(size_t jobnumber)
 {
     if (!is_interactive_now)
@@ -683,6 +729,7 @@ void notify_signaled_job(size_t jobnumber)
     switch (sig) {
 	case SIGINT:
 	    fputc('\n', stderr);
+	    set_interrupted();
 	    break;
 	case SIGPIPE:
 	    break;
@@ -984,6 +1031,10 @@ int continue_job(size_t jobnumber, job_T *job, bool fg)
     if (name != job->j_procs[0].pr_name)
 	free(name);
 
+#if YASH_ENABLE_LINEEDIT && !defined(FG_DONT_SAVE_TERMINAL)
+    bool termsave = fg && le_save_terminal();  /* see below */
+#endif
+
     if (job->j_status != JS_DONE) {
 	if (fg)
 	    put_foreground(job->j_pgid);
@@ -997,6 +1048,11 @@ int continue_job(size_t jobnumber, job_T *job, bool fg)
     int status;
     if (fg) {
 	wait_for_job(jobnumber, true, false, false);
+	put_foreground(shell_pgid);  /* put the shell in the foreground */
+#if YASH_ENABLE_LINEEDIT && !defined(FG_DONT_SAVE_TERMINAL)
+	if (termsave)
+	    le_restore_terminal();
+#endif
 	status = (job->j_status == JS_RUNNING)
 	    ? Exit_SUCCESS : calc_status_of_job(job);
 	if (job->j_status == JS_DONE) {
@@ -1007,6 +1063,18 @@ int continue_job(size_t jobnumber, job_T *job, bool fg)
 	status = (job->j_status == JS_RUNNING) ? Exit_SUCCESS : Exit_FAILURE;
     }
     return status;
+
+    /* We save the terminal state before continuing a job in the foreground and
+     * restore the state after the job has finished. This is because some
+     * programs leave the terminal in the wrong state if they were at first
+     * invoked in the background.
+     * Programs that change the terminal state generally save the state before
+     * changing it and restore it when they are finished. But, if they are
+     * invoked in the background, they save the state that is possibly being
+     * used by another program (typically the shell's line-editing), so the
+     * state that they restore is not the normal state.
+     * The shell tackles this problem by saving and restoring the terminal state
+     * for the continued programs. */
 }
 
 const char fg_help[] = Ngt(
@@ -1078,10 +1146,10 @@ int wait_builtin(int argc, void **argv)
 		    status += TERMSIGOFFSET;
 		    break;
 		}
-		if (job->j_status == JS_DONE) {
+		if (job->j_status != JS_RUNNING) {
 		    if (jobcontrol && is_interactive_now && !posixly_correct)
 			print_job_status(jobnumber, false, false, stdout);
-		    else
+		    else if (job->j_status == JS_DONE)
 			remove_job(jobnumber);
 		}
 	    }
@@ -1097,10 +1165,10 @@ int wait_builtin(int argc, void **argv)
 		    status += TERMSIGOFFSET;
 		    break;
 		}
-		if (job->j_status == JS_DONE) {
+		if (job->j_status != JS_RUNNING) {
 		    if (jobcontrol && is_interactive_now && !posixly_correct)
 			print_job_status(i, false, false, stdout);
-		    else
+		    else if (job->j_status == JS_DONE)
 			remove_job(i);
 		}
 	    }
