@@ -29,6 +29,7 @@
 #include "../sig.h"
 #include "../strbuf.h"
 #include "../util.h"
+#include "../variable.h"
 #include "display.h"
 #include "editing.h"
 #include "key.h"
@@ -41,6 +42,8 @@
 static void reader_init(void);
 static void reader_finalize(void);
 static void read_next(void);
+static int get_read_timeout(void)
+    __attribute__((pure));
 static char pop_prebuffer(void);
 static inline bool has_meta_bit(char c)
     __attribute__((pure));
@@ -163,10 +166,6 @@ static mbstate_t reader_state;
 static xwcsbuf_T reader_second_buffer;
 /* If true, next input will be inserted directly to the main buffer. */
 bool le_next_verbatim;
-/* The time length the shell waits for to see if more input is available when
- * the input sequence is ambiguous.
- * A negative value means no limit. */
-int le_read_timeout = 100;  /* in milliseconds */
 
 /* Initializes the state of the reader.
  * Called for each invocation of `le_readline'. */
@@ -205,7 +204,7 @@ void read_next(void)
     le_display_update();
     fflush(stderr);
     timeout = !wait_for_input(
-	    STDIN_FILENO, true, keycode_ambiguous ? le_read_timeout : -1);
+	    STDIN_FILENO, true, keycode_ambiguous ? get_read_timeout() : -1);
     if (!timeout) {
 	switch (read(STDIN_FILENO, &c, 1)) {
 	    case 0:
@@ -245,17 +244,16 @@ direct_first_buffer:
     keycode_ambiguous = false;
     while (reader_first_buffer.length > 0) {
 	/* check if `reader_first_buffer' is a special sequence */
+	int firstchar = (unsigned char) reader_first_buffer.contents[0];
 	trieget_T tg;
-	if (reader_first_buffer.contents[0] == le_interrupt_char)
+	if (firstchar == le_interrupt_char)
 	    tg = make_trieget(Key_interrupt);
-	else if (reader_first_buffer.contents[0] == le_eof_char)
+	else if (firstchar == le_eof_char)
 	    tg = make_trieget(Key_eof);
-	else if (reader_first_buffer.contents[0] == le_kill_char)
+	else if (firstchar == le_kill_char)
 	    tg = make_trieget(Key_kill);
-	else if (reader_first_buffer.contents[0] == le_erase_char)
+	else if (firstchar == le_erase_char)
 	    tg = make_trieget(Key_erase);
-	else if (reader_first_buffer.contents[0] == '\\')
-	    tg = make_trieget(Key_backslash);
 	else
 	    tg = trie_get(le_keycodes,
 		reader_first_buffer.contents, reader_first_buffer.length);
@@ -345,6 +343,23 @@ process_keymap:
     }
 }
 
+/* Returns a timeout value to be passed to the `wait_for_input' function.
+ * The value is taken from the YASH_LE_TIMEOUT variable. */
+int get_read_timeout(void)
+{
+#ifndef LE_TIMEOUT_DEFAULT
+#define LE_TIMEOUT_DEFAULT 100
+#endif
+
+    const wchar_t *v = getvar(L VAR_YASH_LE_TIMEOUT);
+    if (v != NULL) {
+	int i;
+	if (xwcstoi(v, 0, &i))
+	    return i;
+    }
+    return LE_TIMEOUT_DEFAULT;
+}
+
 /* Appends `s' to the prebuffer. The string `s' is freed in this function. */
 void append_to_prebuffer(char *s)
 {
@@ -398,11 +413,20 @@ trieget_T make_trieget(const wchar_t *keyseq)
 void append_to_second_buffer(wchar_t wc)
 {
     if (le_next_verbatim) {
+	le_next_verbatim = false;
 	le_invoke_command(le_current_mode->default_command, wc);
-    } else if (wc != L'\0') {
-	wb_wccat(&reader_second_buffer, wc);
+    } else {
+	switch (wc) {
+	    case L'\0':
+		break;
+	    case L'\\':
+		wb_cat(&reader_second_buffer, Key_backslash);
+		break;
+	    default:
+		wb_wccat(&reader_second_buffer, wc);
+		break;
+	}
     }
-    le_next_verbatim = false;
 }
 
 
