@@ -223,7 +223,7 @@
 
 /* This flag is set to true when the terminfo database needs to be refreshed
  * because the $TERM variable is changed. */
-_Bool le_need_term_update = true;
+_Bool le_need_term_update = 1;
 
 /* Number of lines and columns in the current terminal. */
 /* Initialized by `le_setupterm'. */
@@ -240,6 +240,9 @@ _Bool le_meta_bit8;
  * Values of entries are `keyseq'. */
 trie_T *le_keycodes = NULL;
 
+/* True if the terminal is set to the keyboard-transmit mode. */
+static _Bool transmit_mode = 0;
+
 
 static inline int is_strcap_valid(const char *s)
     __attribute__((const));
@@ -250,15 +253,8 @@ static _Bool move_cursor_1(char *capone, long count)
     __attribute__((nonnull));
 static _Bool move_cursor_mul(char *capmul, long count, int affcnt)
     __attribute__((nonnull));
-static void print_smkx(void);
-static void print_rmkx(void);
 static int putchar_stderr(int c);
 
-static inline int xtcsetattr(int fd, int opt, const struct termios *term)
-    __attribute__((nonnull));
-static inline int xtcgetattr(int fd, struct termios *term)
-    __attribute__((nonnull));
-
 
 /* Checks if the result of `tigetstr' is valid. */
 int is_strcap_valid(const char *s)
@@ -683,20 +679,26 @@ void le_print_setbg(int color)
     }
 }
 
-/* Prints "smkx" variable. */
+/* Prints "smkx" variable and sets the `transmit_mode' flag. */
 void print_smkx(void)
 {
     char *v = tigetstr(TI_smkx);
-    if (is_strcap_valid(v))
+    if (is_strcap_valid(v)) {
 	tputs(v, 1, putchar_stderr);
+	transmit_mode = 1;
+    }
 }
 
-/* Prints "rmkx" variable. */
+/* Prints "rmkx" variable if the `transmit_mode' flag is set.
+ * The flag is cleared in this function. */
 void print_rmkx(void)
 {
-    char *v = tigetstr(TI_rmkx);
-    if (is_strcap_valid(v))
-	tputs(v, 1, putchar_stderr);
+    if (transmit_mode) {
+	char *v = tigetstr(TI_rmkx);
+	if (is_strcap_valid(v))
+	    tputs(v, 1, putchar_stderr);
+	transmit_mode = 0;
+    }
 }
 
 /* Like `putchar', but prints to `stderr'. */
@@ -721,32 +723,42 @@ void le_alert(void)
 /********** TERMIOS **********/
 
 /* Special characters. */
-char le_eof_char, le_kill_char, le_interrupt_char, le_erase_char;
+int le_eof_char, le_kill_char, le_interrupt_char, le_erase_char;
 
 static struct termios original_terminal_state;
 
+static inline int normchar(cc_t c)
+    __attribute__((const));
+static inline int xtcgetattr(int fd, struct termios *term)
+    __attribute__((nonnull));
+static inline int xtcsetattr(int fd, int opt, const struct termios *term)
+    __attribute__((nonnull));
+
+
 /* Sets the terminal to the "raw" mode.
  * The current state is saved as `original_terminal_state'.
- * `stdin' must be the terminal.
- * Returns true iff successful. */
+ * Returns true iff the terminal (`stdin') has been successfully prepared. */
 _Bool le_set_terminal(void)
 {
     struct termios term;
 
-    if (xtcgetattr(STDIN_FILENO, &term) != 0)
+    /* First we call `tcdrain' to get a clean state and to ensure the shell is
+     * in the foreground. */
+    tcdrain(STDIN_FILENO);
+
+    /* get the original state */
+    if (!le_save_terminal())
 	return 0;
-    original_terminal_state = term;
-    le_eof_char       = TO_CHAR(term.c_cc[VEOF]);
-    le_kill_char      = TO_CHAR(term.c_cc[VKILL]);
-    le_interrupt_char = TO_CHAR(term.c_cc[VINTR]);
-    le_erase_char     = TO_CHAR(term.c_cc[VERASE]);
+    term = original_terminal_state;
+    le_eof_char       = normchar(term.c_cc[VEOF]);
+    le_kill_char      = normchar(term.c_cc[VKILL]);
+    le_interrupt_char = normchar(term.c_cc[VINTR]);
+    le_erase_char     = normchar(term.c_cc[VERASE]);
 
     /* set attributes */
-    term.c_iflag &= ~(IGNBRK | BRKINT | INPCK | ISTRIP
-                     | INLCR | IGNCR | ICRNL | IXON);
+    term.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | INLCR | IGNCR | ICRNL | IXON);
+    term.c_iflag |= IGNPAR;
     term.c_lflag &= ~(ISIG | ECHO | ICANON | IEXTEN);
-    term.c_cflag &= ~(CSIZE | PARENB);
-    term.c_cflag |= CS8;
     term.c_cc[VTIME] = 0;
     term.c_cc[VMIN] = 0;
     if (xtcsetattr(STDIN_FILENO, TCSADRAIN, &term) != 0)
@@ -755,16 +767,13 @@ _Bool le_set_terminal(void)
     /* check if the attributes are properly set */
     if (xtcgetattr(STDIN_FILENO, &term) != 0)
 	goto fail;
-    if ((term.c_iflag & (IGNBRK | BRKINT | INPCK | ISTRIP
-		    | INLCR | IGNCR | ICRNL | IXON))
-	    || (term.c_lflag & (ISIG | ECHO | ICANON | IEXTEN))
-	    || (term.c_cflag & PARENB)
-	    || ((term.c_cflag & CSIZE) != CS8)
+    if ((term.c_iflag & (INLCR | ICRNL))
+	    || (term.c_lflag & (ECHO | ICANON))
 	    || (term.c_cc[VTIME] != 0)
 	    || (term.c_cc[VMIN] != 0))
 	goto fail;
 
-    // XXX it should be configurable whether we print smkx/rmkx or not.
+    // XXX it should be configurable whether we print smkx or not.
     print_smkx();
 
     return 1;
@@ -774,25 +783,31 @@ fail:
     return 0;
 }
 
-/* Restores the terminal to the original state.
- * `stdin' must be the terminal.
- * Returns true iff successful. */
+int normchar(cc_t c)
+{
+    if (c == _POSIX_VDISABLE)
+	return -1;
+    else
+	return (unsigned char) c;
+}
+
+/* Saves the current terminal state in `original_terminal_state'.
+ * Returns true iff `stdin' is a terminal and the terminal state has been
+ * successfully saved. */
+_Bool le_save_terminal(void)
+{
+    return xtcgetattr(STDIN_FILENO, &original_terminal_state) >= 0;
+}
+
+/* Restores the terminal to the original state saved in
+ * `original_terminal_state'.
+ * Returns true iff `stdin' is a terminal and the terminal state has been
+ * successfully restored. */
 _Bool le_restore_terminal(void)
 {
     print_rmkx();
     fflush(stderr);
-    return xtcsetattr(STDIN_FILENO, TCSADRAIN, &original_terminal_state) == 0;
-}
-
-/* Calls `tcsetattr'.
- * If it returns `EINTR' error, re-calls it. */
-int xtcsetattr(int fd, int opt, const struct termios *term)
-{
-    int result;
-    do
-	result = tcsetattr(fd, opt, term);
-    while (result != 0 && errno == EINTR);
-    return result;
+    return xtcsetattr(STDIN_FILENO, TCSADRAIN, &original_terminal_state) >= 0;
 }
 
 /* Calls `tcgetattr'.
@@ -806,5 +821,16 @@ int xtcgetattr(int fd, struct termios *term)
     return result;
 }
 
+/* Calls `tcsetattr'.
+ * If it returns `EINTR' error, re-calls it. */
+int xtcsetattr(int fd, int opt, const struct termios *term)
+{
+    int result;
+    do
+	result = tcsetattr(fd, opt, term);
+    while (result != 0 && errno == EINTR);
+    return result;
+}
+
 
 /* vim: set ts=8 sts=4 sw=4 noet tw=80: */
