@@ -31,7 +31,9 @@
 #if HAVE_GETTEXT
 # include <libintl.h>
 #endif
-#include "alias.h"
+#if YASH_ENABLE_ALIAS
+# include "alias.h"
+#endif
 #include "expand.h"
 #include "input.h"
 #include "option.h"
@@ -328,6 +330,7 @@ struct parsestate_T {
     size_t cindex;
     struct plist_T pending_heredocs;
 #if YASH_ENABLE_ALIAS
+    bool alias_enabled, reparse_alias;
     struct aliaslist_T *caliases;
 #endif
 };
@@ -381,10 +384,10 @@ static void **parse_words_to_paren(void)
     __attribute__((malloc,warn_unused_result));
 static redir_T *tryparse_redirect(void)
     __attribute__((malloc,warn_unused_result));
-static wordunit_T *parse_word_to(aliastype_T type, bool testfunc(wchar_t c))
-    __attribute__((malloc,warn_unused_result,nonnull));
 static inline wordunit_T *parse_word(aliastype_T type)
     __attribute__((malloc,warn_unused_result));
+static wordunit_T *parse_word_to(bool testfunc(wchar_t c))
+    __attribute__((malloc,warn_unused_result,nonnull));
 static void skip_to_next_single_quote(void);
 static wordunit_T *parse_special_word_unit(void)
     __attribute__((malloc,warn_unused_result));
@@ -447,6 +450,11 @@ static size_t cindex;
 /* list of here-documents whose contents are to be read */
 static plist_T pending_heredocs;
 #if YASH_ENABLE_ALIAS
+/* If true, alias substitution is performed. */
+static bool alias_enabled;
+/* If true, the current word is to be re-parsed as alias substitution has been
+ * performed. */
+static bool reparse_alias;
 /* list of currently expanded aliases */
 static struct aliaslist_T *caliases;
 #endif
@@ -466,6 +474,8 @@ struct parsestate_T *save_parse_state(void)
 	.cindex = cindex,
 	.pending_heredocs = pending_heredocs,
 #if YASH_ENABLE_ALIAS
+	.alias_enabled = alias_enabled,
+	.reparse_alias = reparse_alias,
 	.caliases = caliases,
 #endif
     };
@@ -481,6 +491,8 @@ void restore_parse_state(struct parsestate_T *state)
     cindex = state->cindex;
     pending_heredocs = state->pending_heredocs;
 #if YASH_ENABLE_ALIAS
+    alias_enabled = state->alias_enabled;
+    reparse_alias = state->reparse_alias;
     caliases = state->caliases;
 #endif
     free(state);
@@ -527,6 +539,8 @@ int read_and_parse(parseinfo_T *restrict info, and_or_T **restrict result)
     }
     pl_init(&pending_heredocs);
 #if YASH_ENABLE_ALIAS
+    alias_enabled = true;
+    reparse_alias = false;
     caliases = new_aliaslist();
 #endif
 
@@ -804,6 +818,12 @@ and_or_T *parse_command_list(void)
 	    *lastp = ao;
 	    lastp = &ao->next;
 	}
+#if YASH_ENABLE_ALIAS
+	if (reparse_alias) {
+	    assert(!ao);
+	    continue;
+	}
+#endif
 
 	separator = false;
 	ensure_buffer(2);
@@ -814,6 +834,9 @@ and_or_T *parse_command_list(void)
 	    separator = true;
 	}
     }
+#if YASH_ENABLE_ALIAS
+    reparse_alias = false;
+#endif
     return first;
 }
 
@@ -841,6 +864,12 @@ and_or_T *parse_compound_list(void)
 	    *lastp = ao;
 	    lastp = &ao->next;
 	}
+#if YASH_ENABLE_ALIAS
+	if (reparse_alias) {
+	    assert(!ao);
+	    continue;
+	}
+#endif
 
 	separator = false;
 	ensure_buffer(2);
@@ -852,6 +881,9 @@ and_or_T *parse_compound_list(void)
 	}
     }
     cerror |= savecerror;
+#if YASH_ENABLE_ALIAS
+    reparse_alias = false;
+#endif
     return first;
 }
 
@@ -860,9 +892,17 @@ and_or_T *parse_compound_list(void)
  * delimiter "&" or ";" when the function returns. */
 and_or_T *parse_and_or_list(void)
 {
+    pipeline_T *p = parse_pipelines_in_and_or();
+#if YASH_ENABLE_ALIAS
+    if (reparse_alias) {
+	assert(!p);
+	return NULL;
+    }
+#endif
+
     and_or_T *result = xmalloc(sizeof *result);
     result->next = NULL;
-    result->ao_pipelines = parse_pipelines_in_and_or();
+    result->ao_pipelines = p;
     result->ao_async = (cbuf.contents[cindex] == L'&');
     return result;
 }
@@ -871,7 +911,7 @@ and_or_T *parse_and_or_list(void)
 pipeline_T *parse_pipelines_in_and_or(void)
 {
     pipeline_T *first = NULL, **lastp = &first;
-    bool cond = cond;
+    bool cond = false;
 
     for (;;) {
 	pipeline_T *p = parse_pipeline();
@@ -880,6 +920,15 @@ pipeline_T *parse_pipelines_in_and_or(void)
 	    *lastp = p;
 	    lastp = &p->next;
 	}
+#if YASH_ENABLE_ALIAS
+	if (reparse_alias) {
+	    assert(!p);
+	    if (first)
+		goto next;
+	    else
+		break;
+	}
+#endif
 
 	ensure_buffer(2);
 	if (cbuf.contents[cindex] == L'&'
@@ -892,6 +941,7 @@ pipeline_T *parse_pipelines_in_and_or(void)
 	    break;
 	}
 	cindex += 2;
+next:
 	skip_to_next_token();
     }
     return first;
@@ -900,20 +950,38 @@ pipeline_T *parse_pipelines_in_and_or(void)
 /* Parses one pipeline. */
 pipeline_T *parse_pipeline(void)
 {
-    pipeline_T *result = xmalloc(sizeof *result);
-    result->next = NULL;
-    result->pl_cond = false;
+    bool neg;
+    command_T *c;
 
     ensure_buffer(2);
     if (is_token_at(L"!", cindex)) {
-	result->pl_neg = true;
+	neg = true;
 	cindex++;
-	skip_blanks_and_comment();
+#if YASH_ENABLE_ALIAS
+	do {
+#endif
+	    skip_blanks_and_comment();
+	    c = parse_commands_in_pipeline();
+#if YASH_ENABLE_ALIAS
+	    assert(!reparse_alias || !c);
+	} while (reparse_alias);
+#endif
     } else {
-	result->pl_neg = false;
+	neg = false;
+	c = parse_commands_in_pipeline();
+#if YASH_ENABLE_ALIAS
+	if (reparse_alias) {
+	    assert(!c);
+	    return NULL;
+	}
+#endif
     }
 
-    result->pl_commands = parse_commands_in_pipeline();
+    pipeline_T *result = xmalloc(sizeof *result);
+    result->next = NULL;
+    result->pl_commands = c;
+    result->pl_neg = neg;
+    result->pl_cond = false;
     return result;
 }
 
@@ -928,21 +996,37 @@ command_T *parse_commands_in_pipeline(void)
 	    *lastp = c;
 	    lastp = &c->next;
 	}
+#if YASH_ENABLE_ALIAS
+	if (reparse_alias) {
+	    assert(!c);
+	    if (first)
+		goto next;
+	    else
+		break;
+	}
+#endif
 
 	ensure_buffer(2);
 	if (cbuf.contents[cindex] == L'|' && cbuf.contents[cindex+1] != L'|') {
 	    cindex++;
-	    skip_to_next_token();
 	} else {
 	    break;
 	}
+next:
+	skip_to_next_token();
     }
     return first;
 }
 
-/* Parses one command. */
+/* Parses one command.
+ * If the first word is alias-substituted, the `reparse_alias' flag is set and
+ * NULL is returned. */
 command_T *parse_command(void)
 {
+#if YASH_ENABLE_ALIAS
+    reparse_alias = false;
+#endif
+
     /* Note: `check_closing_token_at` includes `ensure_buffer(5)'. */
     const wchar_t *t = check_closing_token_at(cindex);
     if (t) {
@@ -969,7 +1053,17 @@ command_T *parse_command(void)
     if (t)
 	return parse_compound_command(t);
 
-    unsigned long lineno = cinfo->lineno;
+#if YASH_ENABLE_ALIAS
+    if (alias_enabled) {
+	size_t len = count_name_length(is_alias_name_char);
+	substaliasflags_T flags = AF_NONGLOBAL | AF_NORECUR;
+	if (substitute_alias(&cbuf, cindex, len, caliases, flags)) {
+	    reparse_alias = true;
+	    return NULL;
+	}
+    }
+#endif
+
     command_T *result = tryparse_function();
     if (result)
 	return result;
@@ -980,7 +1074,7 @@ command_T *parse_command(void)
 
     result->next = NULL;
     result->refcount = 1;
-    result->c_lineno = lineno;
+    result->c_lineno = cinfo->lineno;
     result->c_type = CT_SIMPLE;
     redirlastp = parse_assignments_and_redirects(result);
     result->c_words = parse_words_and_redirects(redirlastp, true);
@@ -992,6 +1086,8 @@ command_T *parse_command(void)
 }
 
 /* Parses assignments and redirections.
+ * Tokens but the first one are subject to any-type alias substitution,
+ * including the word just after the parsed assignments and redirections.
  * The first following word is subject to any-type alias substitution.
  * The results are assigned to `c->c_assigns' and `c->c_redirs'.
  * The new `redirlastp' is returned. */
@@ -1006,11 +1102,6 @@ redir_T **parse_assignments_and_redirects(command_T *c)
     c->c_redirs = NULL;
     while (ensure_buffer(1),
 	    !is_command_delimiter_char(cbuf.contents[cindex])) {
-#if YASH_ENABLE_ALIAS
-	size_t len = count_name_length(is_alias_name_char);
-	substitute_alias(&cbuf, cindex, len, caliases, false);
-	skip_blanks_and_comment();
-#endif
 	if ((redir = tryparse_redirect())) {
 	    *redirlastp = redir;
 	    redirlastp = &redir->next;
@@ -1020,6 +1111,13 @@ redir_T **parse_assignments_and_redirects(command_T *c)
 	} else {
 	    break;
 	}
+#if YASH_ENABLE_ALIAS
+	if (alias_enabled) {
+	    size_t len = count_name_length(is_alias_name_char);
+	    substitute_alias(&cbuf, cindex, len, caliases, AF_NONGLOBAL);
+	    skip_blanks_and_comment();
+	}
+#endif
     }
     return redirlastp;
 }
@@ -1040,9 +1138,9 @@ void **parse_words_and_redirects(redir_T **redirlastp, bool first)
     while (ensure_buffer(1),
 	    !is_command_delimiter_char(cbuf.contents[cindex])) {
 #if YASH_ENABLE_ALIAS
-	if (!first) {
+	if (!first && alias_enabled) {
 	    size_t len = count_name_length(is_alias_name_char);
-	    substitute_alias(&cbuf, cindex, len, caliases, true);
+	    substitute_alias(&cbuf, cindex, len, caliases, 0);
 	    skip_blanks_and_comment();
 	}
 #endif
@@ -1067,9 +1165,9 @@ void parse_redirect_list(redir_T **lastp)
 {
     for (;;) {
 #if YASH_ENABLE_ALIAS
-	if (!posixly_correct) {
+	if (!posixly_correct && alias_enabled) {
 	    size_t len = count_name_length(is_alias_name_char);
-	    substitute_alias(&cbuf, cindex, len, caliases, true);
+	    substitute_alias(&cbuf, cindex, len, caliases, 0);
 	}
 #endif
 
@@ -1245,35 +1343,39 @@ parse_command:
     return result;
 }
 
+/* Expands an alias and parses a word at the current index.
+ * `type' specifies the type of aliases to be expanded.  */
 wordunit_T *parse_word(aliastype_T type)
-{
-    return parse_word_to(type, is_token_delimiter_char);
-}
-
-/* Expands an alias and parses a word at the current index.
- * `type' specifies the type of aliases to be expanded.
- * `testfunc' is a function that determines if a character is a word delimiter.
- * It must return true for L'\0'.
- * The parsing process proceeds up to an unescaped character `testfunc' returns
- * false for. It is not an error if there is no characters to be a word, in
- * which case NULL is returned. */
-wordunit_T *parse_word_to(aliastype_T type, bool testfunc(wchar_t c))
 {
 #if YASH_ENABLE_ALIAS
-    switch (type) {
+    if (alias_enabled) {
+	switch (type) {
 	case noalias:
 	    break;
 	case globalonly:
 	case anyalias:;
 	    size_t len = count_name_length(is_alias_name_char);
-	    substitute_alias(&cbuf, cindex, len, caliases, type == globalonly);
+	    substaliasflags_T flags = type == globalonly ? 0 : AF_NONGLOBAL;
+	    substitute_alias(&cbuf, cindex, len, caliases, flags);
 	    skip_blanks_and_comment();
 	    break;
+	}
     }
 #else /* !YASH_ENABLE_ALIAS */
     (void) type;
 #endif
 
+    return parse_word_to(is_token_delimiter_char);
+}
+
+/* Parses a word at the current index.
+ * `testfunc' is a function that determines if a character is a word delimiter.
+ * It must return true for L'\0'.
+ * The parsing process proceeds up to an unescaped character `testfunc' returns
+ * false for. It is not an error if there is no characters to be a word, in
+ * which case NULL is returned. */
+wordunit_T *parse_word_to(bool testfunc(wchar_t c))
+{
     wordunit_T *first = NULL, **lastp = &first, *wu;
     bool indq = false;  /* in double quotes? */
     size_t startindex = cindex;
@@ -1518,12 +1620,12 @@ make_name:
     ensure_buffer(3);
     if (!posixly_correct && cbuf.contents[cindex] == L'[') {
 	cindex++;
-	pe->pe_start = parse_word_to(noalias, is_comma_or_closing_bracket);
+	pe->pe_start = parse_word_to(is_comma_or_closing_bracket);
 	if (!pe->pe_start)
 	    serror(Ngt("index missing"));
 	if (cbuf.contents[cindex] == L',') {
 	    cindex++;
-	    pe->pe_end = parse_word_to(noalias, is_comma_or_closing_bracket);
+	    pe->pe_end = parse_word_to(is_comma_or_closing_bracket);
 	    if (!pe->pe_end)
 		serror(Ngt("index missing"));
 	}
@@ -1592,10 +1694,10 @@ parse_match:
 	cindex += 1;
     }
     if ((pe->pe_type & PT_MASK) == PT_MATCH) {
-	pe->pe_match = parse_word_to(noalias, is_closing_brace);
+	pe->pe_match = parse_word_to(is_closing_brace);
 	goto check_closing_paren_and_finish;
     } else {
-	pe->pe_match = parse_word_to(noalias, is_slash_or_closing_brace);
+	pe->pe_match = parse_word_to(is_slash_or_closing_brace);
     }
 
     ensure_buffer(1);
@@ -1603,7 +1705,7 @@ parse_match:
 	goto check_closing_paren_and_finish;
 parse_subst:
     cindex++;
-    pe->pe_subst = parse_word_to(noalias, is_closing_brace);
+    pe->pe_subst = parse_word_to(is_closing_brace);
 
 check_closing_paren_and_finish:
     ensure_buffer(1);
@@ -2398,7 +2500,7 @@ void print_errmsg_token_missing(const wchar_t *t, size_t index)
 }
 
 
-/********** Functions that Convert Parse Trees into Strings**********/
+/********** Functions that Convert Parse Trees into Strings **********/
 
 static void print_and_or_lists(
 	xwcsbuf_T *restrict buf, const and_or_T *restrict andors,
@@ -2520,9 +2622,13 @@ void print_command_content(xwcsbuf_T *restrict buf, const command_T *restrict c)
 	break;
     case CT_SUBSHELL:
 	wb_wccat(buf, L'(');
-	print_and_or_lists(buf, c->c_subcmds, true);
-	assert(iswblank(buf->contents[buf->length - 1]));
-	wb_insert(buf, buf->length - 1, L")");
+	if (c->c_subcmds) {
+	    print_and_or_lists(buf, c->c_subcmds, true);
+	    assert(iswblank(buf->contents[buf->length - 1]));
+	    wb_insert(buf, buf->length - 1, L")");
+	} else {
+	    wb_wccat(buf, L')');
+	}
 	break;
     case CT_IF:
 	wb_cat(buf, L"if ");
