@@ -235,23 +235,25 @@ int copy_as_shellfd(int fd)
 /* Duplicates the underlining file descriptor of the specified stream.
  * The original stream is closed whether successful or not.
  * A new stream is open with the new FD using `fdopen' and returned.
- * The new stream's underlining file descriptor is registered as a shell FD and
- * set to non-blocking.
+ * The new stream's underlining file descriptor is registered as a shell FD and,
+ * if `nonblock' is true, set to non-blocking.
  * If NULL is given, this function just returns NULL without doing anything. */
-FILE *reopen_with_shellfd(FILE *f, const char *mode)
+FILE *reopen_with_shellfd(FILE *f, const char *mode, bool nonblock)
 {
     if (!f)
 	return NULL;
 
     int newfd = copy_as_shellfd(fileno(f));
     fclose(f);
-    if (newfd < 0 || !set_nonblocking(newfd))
+    if (newfd < 0 || (nonblock && !set_nonblocking(newfd)))
 	return NULL;
     else
 	return fdopen(newfd, mode);
 }
 
-/* Opens `ttyfd' */
+/* Opens `ttyfd'.
+ * On failure, an error message is printed and `do_job_control' is set to false.
+ */
 void open_ttyfd(void)
 {
     if (ttyfd < 0) {
@@ -279,6 +281,8 @@ struct savefd_T {
     bool sf_stdin_redirected;  /* original `is_stdin_redirected' */
 };
 
+static char *expand_redir_filename(const struct wordunit_T *filename)
+    __attribute__((malloc,warn_unused_result));
 static void save_fd(int oldfd, savefd_T **save);
 static int open_file(const char *path, int oflag)
     __attribute__((nonnull));
@@ -314,12 +318,16 @@ bool open_redirections(const redir_T *r, savefd_T **save)
 	}
 
 	/* expand rd_filename */
+#ifdef NDEBUG
+	char *filename;
+#else
 	char *filename = filename;
+#endif
 	switch (r->rd_type) {
 	    case RT_INPUT:  case RT_OUTPUT:  case RT_CLOBBER:  case RT_APPEND:
 	    case RT_INOUT:  case RT_DUPIN:   case RT_DUPOUT:   case RT_PIPE:
 	    case RT_HERESTR:
-		filename = expand_single_with_glob(r->rd_filename, tt_single);
+		filename = expand_redir_filename(r->rd_filename);
 		if (!filename)
 		    return false;
 		break;
@@ -419,6 +427,22 @@ openwithflags:
     return true;
 }
 
+/* Expands the filename for a redirection. */
+char *expand_redir_filename(const struct wordunit_T *filename)
+{
+    if (is_interactive) {
+	return expand_single_with_glob(filename, tt_single);
+    } else {
+	wchar_t *result = expand_single(filename, tt_single);
+	if (result == NULL)
+	    return NULL;
+	char *mbsresult = realloc_wcstombs(unescapefree(result));
+	if (!mbsresult)
+	    xerror(EILSEQ, Ngt("redirection"));
+	return mbsresult;
+    }
+}
+
 /* Saves the specified file descriptor if `save' is non-NULL. */
 void save_fd(int fd, savefd_T **save)
 {
