@@ -63,9 +63,10 @@ static bool input_is_interactive_terminal(const parseinfo_T *pinfo)
 
 /* the process ID of the shell */
 pid_t shell_pid;
-/* the initial/current process group ID of the shell */
-static pid_t initial_pgid;
+/* the process group ID of the shell */
 pid_t shell_pgid;
+/* In a job-controlled subshell, `shell_pgid' is set to the subshell's pgid.
+ * `shell_pid' is always the process ID of the main shell process. */
 
 /* an equivalent to the -f flag for "exit" builtin */
 static bool forceexit;
@@ -206,10 +207,10 @@ int main(int argc, char **argv)
     if (wargv[xoptind] && wcscmp(wargv[xoptind], L"-") == 0)
 	xoptind++;
 
-    /* `shell_pid' must be initialized after the options have been parsed.
+    /* `shell_pgid' must be initialized after the options have been parsed.
      * This is required for the `set_monitor_option' function to work. */
     shell_pid = getpid();
-    initial_pgid = shell_pgid = getpgrp();
+    shell_pgid = getpgrp();
     init_cmdhash();
     init_homedirhash();
     init_variables();
@@ -221,9 +222,15 @@ int main(int argc, char **argv)
     init_alias();
 #endif
 
+#ifdef NDEBUG
+    wchar_t *command;
+    FILE *input;
+    const char *inputname;
+#else
     wchar_t *command = command;
     FILE *input = input;
     const char *inputname = inputname;
+#endif
 
     if (shopt_read_arg && shopt_read_stdin) {
 	xerror(0, Ngt("both of -c and -s options cannot be used at once"));
@@ -249,7 +256,7 @@ int main(int argc, char **argv)
 	} else {
 	    inputname = argv[xoptind];
 	    input = fopen(inputname, "r");
-	    input = reopen_with_shellfd(input, "r");
+	    input = reopen_with_shellfd(input, "r", true);
 	    command_name = wargv[xoptind++];
 	    if (!input) {
 		int errno_ = errno;
@@ -270,10 +277,11 @@ int main(int argc, char **argv)
     is_interactive_now = is_interactive;
     if (!do_job_control_set)
 	do_job_control = is_interactive;
-    if (is_interactive || do_job_control)
+    if (do_job_control) {
 	open_ttyfd();
+	ensure_foreground();
+    }
     set_signals();
-    set_own_pgid();
     set_positional_parameters(wargv + xoptind);
     if (getuid() == geteuid() && getgid() == getegid()) {
 	if (!noprofile)
@@ -301,7 +309,7 @@ static void execute_profile(void)
     if (wpath) {
 	char *path = realloc_wcstombs(wpath);
 	if (path) {
-	    FILE *f = reopen_with_shellfd(fopen(path, "r"), "r");
+	    FILE *f = reopen_with_shellfd(fopen(path, "r"), "r", true);
 	    if (f) {
 		exec_input(f, path, false, false);
 		fclose(f);
@@ -341,7 +349,7 @@ static void execute_rcfile(const wchar_t *rcfile)
     }
     if (!path)
 	return;
-    f = reopen_with_shellfd(fopen(path, "r"), "r");
+    f = reopen_with_shellfd(fopen(path, "r"), "r", true);
     if (f) {
 	exec_input(f, path, false, false);
 	fclose(f);
@@ -370,23 +378,9 @@ void exit_shell_with_status(int status)
 	    finalize_history();
 #endif
     }
-    finalize_shell();
     exit(exitstatus);
 }
 
-/* Does what to do before exiting/suspending the shell.
- * The EXIT trap should be executed beforehand if needed. */
-void finalize_shell(void)
-{
-    reset_own_pgid();
-}
-
-/* Cancels `finalize_shell'. */
-void reinitialize_shell(void)
-{
-    set_own_pgid();
-}
-
 /* Prints the help message to stdout */
 void print_help(void)
 {
@@ -456,39 +450,6 @@ void print_version(void)
 }
 
 
-/* If job control is active, sets the process group ID of the shell to its
- * process ID. */
-void set_own_pgid(void)
-{
-    if (doing_job_control_now && initial_pgid > 0 && shell_pgid != shell_pid) {
-	if (setpgid(0, 0) == 0) {
-	    shell_pgid = shell_pid;
-	    put_foreground(shell_pgid);
-	}
-    }
-}
-
-/* If job control is active, resets the process group ID of the shell.
- * The initial process group ID is restored. */
-void reset_own_pgid(void)
-{
-    if (doing_job_control_now && initial_pgid > 0
-	    && initial_pgid != shell_pgid) {
-	if (setpgid(0, initial_pgid) == 0) {
-	    shell_pgid = initial_pgid;
-	    put_foreground(shell_pgid);
-	}
-    }
-}
-
-/* Forgets the value of `initial_pgid' so that `reset_own_pgid' is no longer
- * effective. */
-void forget_initial_pgid(void)
-{
-    initial_pgid = 0;
-}
-
-
 /********** Functions to execute commands **********/
 
 /* Parses a multibyte string and executes the commands.
@@ -570,7 +531,11 @@ void exec_input(FILE *f, const char *name, bool intrinput, bool finally_exit)
  * If there are no commands in `code', `laststatus' is set to 0. */
 void parse_and_exec(parseinfo_T *pinfo, bool finally_exit)
 {
+#ifdef NDEBUG
+    struct execinfo *ei;
+#else
     struct execinfo *ei = ei;
+#endif
     if (!finally_exit)
 	ei = save_execinfo();
 
@@ -582,19 +547,23 @@ void parse_and_exec(parseinfo_T *pinfo, bool finally_exit)
 	if (pinfo->intrinput) {
 	    forceexit = nextforceexit;
 	    nextforceexit = false;
-	}
-	if (return_pending()) {
-	    if (pinfo->intrinput) {
-		reset_execinfo();
+	    if (!is_interrupted() && return_pending()) {
 		xerror(0, Ngt("return: not in function or sourced file"));
 		laststatus = Exit_FAILURE;
-	    } else if (finally_exit) {
-		wchar_t argv0[] = L"return";
-		reset_execinfo();
-		exit_builtin(1, (void *[]) { argv0, NULL });
-		continue;
-	    } else
-		goto out;
+	    }
+	    reset_sigint();
+	    reset_execinfo();
+	} else {
+	    if (return_pending()) {
+		if (finally_exit) {
+		    wchar_t argv0[] = L"return";
+		    reset_execinfo();
+		    exit_builtin(1, (void *[]) { argv0, NULL });
+		    continue;
+		} else {
+		    goto out;
+		}
+	    }
 	}
 
 	switch (read_and_parse(pinfo, &commands)) {
@@ -724,43 +693,50 @@ const char exit_help[] = Ngt(
 "`exit' twice in a row to avoid the warning and really exit.\n"
 );
 
-/* The "suspend" builtin */
-/* This builtin accepts the -f (--force) option, which is silently ignored for
- * bash-compatibility. Yash's "suspend" builtin always suspends the shell
- * whether or not it is a login shell. */
+/* The "suspend" builtin, which accepts the following options:
+ *  -f: suspend even if it may cause a deadlock. */
 int suspend_builtin(int argc, void **argv)
 {
+    bool force = false;
     wchar_t opt;
 
     xoptind = 0, xopterr = true;
     while ((opt = xgetopt_long(argv, L"f", force_help_options, NULL))) {
 	switch (opt) {
 	    case L'f':
+		force = true;
 		break;
 	    case L'-':
 		return print_builtin_help(ARGV(0));
 	    default:  print_usage:
-		fprintf(stderr, gt("Usage:  suspend\n"));
+		fprintf(stderr, gt("Usage:  suspend [-f]\n"));
 		return Exit_ERROR;
 	}
     }
     if (argc != xoptind)
 	goto print_usage;
+    if (!force && is_interactive_now && getsid(0) == shell_pgid) {
+	xerror(0, Ngt("refusing to suspend because of possible deadlock.\n"
+		    "Use the -f option to suspend anyway."));
+	return Exit_FAILURE;
+    }
 
     bool ok;
 
-    finalize_shell();
     ok = send_sigstop_to_myself();
     if (!ok)
 	xerror(errno, Ngt("cannot send SIGSTOP signal"));
-    reinitialize_shell();
+    if (doing_job_control_now)
+	ensure_foreground();
     return ok ? Exit_SUCCESS : Exit_FAILURE;
 }
 
 const char suspend_help[] = Ngt(
 "suspend - suspend shell\n"
-"\tsuspend\n"
-"Suspends the shell until it receives a SIGCONT signal.\n"
+"\tsuspend [-f]\n"
+"Suspends the shell until it receives SIGCONT.\n"
+"This command refuses to suspend the shell if it is interactive and is a\n"
+"session leader.\n"
 );
 
 
